

#######################################################################
#######################################################################
#######################################################################


library(Hmisc)


################################################################################
### This function was copied from the R package reflimR to compute the       ###
### Bowley skewness required for the function def.lognorm.                   ###
################################################################################
bowley <- function(x, alpha=0.25, digits=3){
  #expects a numeric vector x
  #returns the Bowley skewness of a range between alpha and 1-alpha
  xx <- as.numeric(as.character(x)) #introduces NA for non-numeric values
  xx <- xx[!is.na(xx)] #removes NAs
  if (length(xx) < 10){
    print("bowley() error: The minimum number of values is 10.")
    return(NULL)
  }
  q1 <- quantile(xx, alpha)
  q2 <- quantile(xx, 0.5)
  q3 <- quantile(xx, 1-alpha)
  bs <- (q1 - 2 * q2 + q3) / (q3 - q1)
  return(round(bs, digits))
} #bowley


################################################################################
### This function was copied from the R package reflimR to decide between    ###
### normal or log-normal distribution.                                       ###
################################################################################
def.lognorm <- function(x, cutoff = 0.05, alpha = 0.25, plot.it = TRUE, 
                        main = "", xlab = "x", ylab = "density", 
                        add.tx = "distribution", setvalues = NULL){
  #expects a numeric vector x > 0
  #compares the Bowley skewness for x and log(x)
  #returns TRUE, if a lognormal distribution should be assumed
  #optional: plots density of x, indicates distribution type plus additional text as a legend
  #optional: draws vertical lines for RI setvalues from LIS, e. g. c(10, 35)
  xx <- x + 1  #shift avoids numbers < 1
  s <- c(NA,NA)
  s[1] <- bowley(xx, alpha=alpha)
  s[2] <- bowley(log(xx), alpha=alpha)
  if(s[1] < 0){logtype <- FALSE}
  else{
    logtype <- (s[1] - s[2]) >= cutoff
  }
  if (plot.it){
    d <- density(x)
    hist(x, freq = FALSE, ylim = c(0, max(d$y) * 1.2), main = main, xlab = xlab, ylab = ylab, col = "lightgrey")
    lines(d$x, d$y)
    grid()
    if (!is.null(setvalues)){abline(v = setvalues, lty = 2)}
    legend("topright", legend = ifelse(logtype, paste("lognormal", add.tx), paste("normal", add.tx)))
  }
  return(list("lognormal" = logtype, "BowleySkewness" = s))
} #def.lognorm



################################################################################
### Functions for Cardano's formula.                                         ###
### Copied from:                                                             ###
################################################################################
### https://stackoverflow.com/questions/22598722/how-to-solve-cubic-equation-analytically-exact-solution-in-r
# -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -
# Return all  the complex cube roots of a number
cuberoot <- function(x){
  return( as.complex(x)^(1/3)*exp(c(0,2,4)*1i*pi/3) );
}
# -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -
# cubsol solves analytically the cubic equation and
# returns a list whose first element is the real roots and the
# second element the complex roots.
# test with :
#a = -1; b=-10; c=0; d=50; x=0.01*(-1000:1500); plot(x,a*x^3+b*x^2+c*x+d,t='l'); abline(h=0)
# coefs = c(a,b,c,d)
cubsol <- function(coeffs) {
  if (!(length(coeffs) == 4)){
    stop('Please provide cubsol with a 4-vector of coefficients')
  }
  a = coeffs[1]; b=coeffs[2]; c=coeffs[3]; d=coeffs[4];
  rts = list();  

  p <- -b/3/a
  q <- p^3 + (b*c-3*a*(d))/(6*a^2)
  r <- c/3/a

  s0 = q^2+(r-p^2)^3; 
  xtemp = as.complex(rep(0,9));    
  if (s0 >= 0){ nReRts=1; } else {nReRts=3; }
  # Now find all the roots in complex space:
  s0 = as.complex(s0);
  s1 = cuberoot(q+s0^0.5)
  s2 = cuberoot(q-s0^0.5);
  xtemp[1:3] <- s1+ s2 +p;  # I think this is meant to always contain
                            # the sure real soln.
  # Second and third solution;
  iSqr3 = sqrt(3)*1i; 
  xtemp[4:6] = p - 0.5*(s1+s2 + iSqr3*(s1-s2));
  xtemp[7:9] = p - 0.5*(s1+s2 - iSqr3*(s1-s2));
  ind1 = which.min(abs(a*xtemp[1:3]^3 + b*xtemp[1:3]^2 +c*xtemp[1:3] +d))
  ind2 = 3+which.min(abs(a*xtemp[4:6]^3 + b*xtemp[4:6]^2 +c*xtemp[4:6] +d))
  ind3 = 6+which.min(abs(a*xtemp[7:9]^3 + b*xtemp[7:9]^2 +c*xtemp[7:9] +d))

  if (nReRts == 1){  
    rts[[1]] = c(Re(xtemp[ind1])); 
    rts[[2]] = xtemp[c(ind2,ind3)]
  } else {  # three real roots
    rts[[1]] = Re(xtemp[c(ind1,ind2,ind3)]); 
    rts[[2]] = numeric();
  }
  return(rts)

} # end of function cubsol


################################################################################
### Computes the coefficients for the cubic Eq. (15).                        ###
###                                                                          ###
### Arguments                                                                ###
### x      The array of x_j values in Eq. (15).                              ###
### delta  The delta value in Eq. (15).                                      ### 
### alpha  The alpha value in Eq. (15).                                      ### 
###                                                                          ###
### Value                                                                    ###
### Array with the coefficients for x^3,...,x^0                              ###
################################################################################
compute.deriv.coefs <- function(x,delta,alpha){
  
  a <- alpha
  d <- delta
  
  sumx <- sum(x)
  sumx2 <- sum(x^2)
  sumx3 <- sum(x^3)
 
  c0 <- sumx3 - a*sumx - d*sumx 
  c1 <- a*length(x) + d*length(x) - 3*sumx2
  c2 <- 3*sumx
  c3 <- -length(x)

  return(c(c3,c2,c1,c0))  
}


################################################################################
### Same as compute.deriv.coefs but with pre-computed sums of powers of the  ###
### x_j values.                                                              ###
###                                                                          ###
### Arguments                                                                ###
### x      The array of x_j values in Eq. (15).                              ###
### delta  The delta value in Eq. (15).                                      ### 
### alpha  The alpha value in Eq. (15).                                      ### 
### sumx   sum(x)                                                            ###
### sumx2  sum(x^2)                                                          ###
### sumx3  sum(x^3)                                                          ###
###                                                                          ###
### Value                                                                    ###
### Array with the coefficients for x^3,...,x^0                              ###
################################################################################
compute.deriv.coefs2 <- function(x,delta,alpha,sumx,sumx2,sumx3){
  
  a <- alpha
  d <- delta
 
  c0 <- sumx3 - a*sumx - d*sumx 
  c1 <- a*length(x) + d*length(x) - 3*sumx2
  c2 <- 3*sumx
  c3 <- -length(x)

  return(c(c3,c2,c1,c0))  
}





################################################################################
### Computes the cluster centre as described in Algorithm 2.                 ###
###                                                                          ###
### Arguments                                                                ###
### x      The array of x_j values.                                          ###
### delta  The delta value in Eq. (15).                                      ### 
### alpha  The alpha value in Eq. (15).                                      ### 
###                                                                          ###
### Value                                                                    ###
### Cluster centre                                                           ###
################################################################################
compute.cluster.centre <- function(x,delta=NULL,alpha=NULL){

  d <- delta
  if (is.null(d)){
    quants <- quantile(x,c(0.1,0.5,0.9))
	d <- max(c(quants[3]-quants[2],quants[2]-quants[1]))
	d <- d^2
  }
  
  a <- alpha
  if (is.null(a)){
    medi <- median(x)
	maxd <- max(c(medi-min(x),max(x)-medi))
    a <- maxd^2 - d
  }
  
  
  
  roots <- cubsol(compute.deriv.coefs(x,d,a))   
  is.found <- (length(roots[[1]])==3)
  while (!is.found){
    a <- 1.1*a
  	roots <- cubsol(compute.deriv.coefs(x,d,a))
    is.found <- (length(roots[[1]])==3)	
  }
  
  
  cluster.centre <- roots[[1]][2]
  
  return(list(cluster.centre=cluster.centre,delta=d,alpha=a))
}


################################################################################
### Same as compute.cluster.centre but with pre-computed sums of powers      ###
### of the x_j values.                                                       ###
###                                                                          ###
### Arguments                                                                ###
### x      The array of x_j values.                                          ###
### sumx   sum(x)                                                            ###
### sumx2  sum(x^2)                                                          ###
### sumx3  sum(x^3)                                                          ###
### delta  The delta value in Eq. (15).                                      ### 
### alpha  The alpha value in Eq. (15).                                      ### 
###                                                                          ###
### Value                                                                    ###
### Cluster centre                                                           ###
################################################################################
compute.cluster.centre2 <- function(x,sumx,sumx2,sumx3,delta=NULL,alpha=NULL){

  d <- delta
  if (is.null(d)){
    quants <- quantile(x,c(0.1,0.5,0.9))
	d <- max(c(quants[3]-quants[2],quants[2]-quants[1]))
	d <- d^2
  }
  
  a <- alpha
  if (is.null(a)){
    medi <- median(x)
	maxd <- max(c(medi-min(x),max(x)-medi))
    a <- maxd^2 - d
  }
  
  
  
  roots <- cubsol(compute.deriv.coefs2(x,d,a,sumx,sumx2,sumx3))   
  is.found <- (length(roots[[1]])==3)
  while (!is.found){
    a <- 1.1*a
  	roots <- cubsol(compute.deriv.coefs2(x,d,a,sumx,sumx2,sumx3))
    is.found <- (length(roots[[1]])==3)	
  }
  
  
  cluster.centre <- roots[[1]][2]
  
  return(list(cluster.centre=cluster.centre,delta=d,alpha=a))
}




################################################################################
### Computes the membership degrees according to Eq. (13).                   ###
###                                                                          ###
### Arguments                                                                ###
### x      The array of x_j values.                                          ###
### cluster.centre  The cluster centre.                                      ###
### delta  The delta value in Eq. (13).                                      ### 
### alpha  The alpha value in Eq. (13).                                      ### 
###                                                                          ###
### Value                                                                    ###
### Cluster centre                                                           ###
################################################################################
compute.membership.degrees <- function(x,cluster.centre,delta=1,alpha=100){
  return(1/2 + (delta - (x - cluster.centre)^2)/(2*alpha))
}



################################################################################
### Cuts off values outside [0,1].                                           ###
###                                                                          ###
### Arguments                                                                ###
### w     Real (floating point) number.                                      ###
###                                                                          ###
### Value                                                                    ###
### Value between 0 and 1.                                                   ###
################################################################################
cut.weights <- function(w){
  if (w<0){return(0)}
  if (w>1){return(1)}
  return(w)
}


################################################################################
### Computes the weighted mean and the weighted standard deviation.          ###
###                                                                          ###
### Arguments                                                                ###
### x     Array of values for which the weighted mean and the weighted       ###
###       standard deviation should be computed.                             ###
### u     Weights (Weights outside [0,1] are cut off at 0 and 1,             ###
###       respectively.)                                                     ###
###                                                                          ###
### Value                                                                    ###
### Weighted mean and weighted standard deviation.                           ###
################################################################################
compute.mean.sd <- function(x,u){
  u.cut <- sapply(u,cut.weights)
  w.mean <- wtd.mean(x,u.cut,normwt=T)
  w.sd <- sqrt(wtd.var(x,u.cut,normwt=T))
  
  return(list(w.mean=w.mean,w.sd=w.sd))
}




################################################################################
### Clusters the data for different delta values and computes the sequence   ###
### of weighted means and weighted standard deviations.                      ###
###                                                                          ###
### Arguments                                                                ###
### x         Array of real (floating point)values.                          ###
### no.steps  No of steps for the reduction of delta.                        ###
### delta.min Smallest delta value.                                          ###
### alpha     The alpha value for the clustering                             ###
###                                                                          ###
### Value                                                                    ###
### Data frame with the weighted means and weighted standard deviations.     ###
################################################################################
mean.sd.seq <- function(x,no.steps=500,delta.min=NULL,alpha=NULL){
  quants <- quantile(x,c(0.1,0.4,0.5,0.6,0.9))
  delta <- (max(c(quants[5]-quants[3],quants[3]-quants[1])))^2
  if (is.null(alpha)){
    alpha <- delta/4
  }
  
  if (is.null(delta.min)){
    delta.min <- (quants[4]-quants[2])^2
  }
  
  sumx <- sum(x)
  sumx2 <- sum(x^2)
  sumx3 <- sum(x^3)
  
  means <- rep(0.1,no.steps)
  sds <- rep(0.1,no.steps)
  
  for (i in 1:no.steps){
	deltai <- delta.min + (delta-delta.min)*(no.steps-i)/no.steps
	cluce <- compute.cluster.centre2(x,sumx,sumx2,sumx3,delta=deltai,alpha=alpha)
	u <- compute.membership.degrees(x,cluce$cluster.centre,delta=cluce$delta,alpha=cluce$alpha)
    msd <- compute.mean.sd(x,u)
	means[i] <- msd$w.mean
	sds[i] <- msd$w.sd
  }
  return(data.frame(means,sds))
}



################################################################################
### Implementation of Algorithm 3.                                           ###
###                                                                          ###
### Arguments                                                                ###
### x         Array of real (floating point)values.                          ###
### no.steps  No of steps for the reduction of delta.                        ###
### delta.min Smallest delta value.                                          ###
### alpha     The alpha value for the clustering                             ###
###                                                                          ###
### Value                                                                    ###
### Data frame with the weighted means and weighted standard deviations.     ###
################################################################################
oneClusterRI <- function(x,no.steps=100,delta.min=NULL,alpha=NULL){
  is.lognorm <- def.lognorm(x,plot.it=F)$lognormal
  xx <- x
  if (is.lognorm){
    xx <- log(x)
  }
   
  mesd <- mean.sd.seq(xx,no.steps=no.steps,delta.min=delta.min,alpha=alpha)
  
  emean <- median(mesd[,1])
  esd <- median(mesd[,2])
  
  ll <- qnorm(0.025,mean=emean,sd=esd)
  ul <- qnorm(0.975,mean=emean,sd=esd)
  
  if (is.lognorm){
    ll <- exp(ll)
	ul <- exp(ul)
  }
  
  return(c(ll,ul))
}
  



#################################################################################
#################################################################################
######################   Example   ##############################################
#################################################################################
#################################################################################




### Generate data that reflect hemoglobin values 
hb <- rnorm(500,mean=14,sd=1) # Healthy individuals
hb2 <- c(hb,rnorm(50,mean=11,sd=1)) # Anaemic individuals


oneClusterRI(hb2)





